"""Parse domain names from provided list of repositories."""

from autogit.data_types import RepoState


def parse_domains(repos: dict[str, RepoState]) -> None:
    pass


# curl --header "PRIVATE-TOKEN: personal-access-token" your-gitlab-url/api/v4/version
